--[[ UPVALUES ]]--
local tinsert, twipe, tremove, abs, sqrt, min, max, floor, random, pow = table.insert, table.wipe, table.remove, math.abs, math.sqrt, math.min, math.max, math.floor, math.random, math.pow

--[[ MODULES ]] --
if not Doodleboard then Doodleboard = {} end
local libmg = LibStub("LibMouseGestures-1.0");

--[[ DEFAULT VARIABLES ]]--
local Doodlepad = {messages = {}};
local myRec,overlaySelector,Storyboard_Frame = nil,nil,nil;
local instanceMaps = {
	{0, "Blank sheet", nil},
	{1, "Blackwing Descent (1)", [[BlackwingDescent\BlackwingDescent1_]]},
	{2, "Blackwing Descent (2)", [[BlackwingDescent\BlackwingDescent2_]]},
	{3, "Baradin Hold", [[BaradinHold\BaradinHold1_]]},
	{4, "Grim Batol", [[GrimBatol\GrimBatol1_]]},
	{5, "Halls Of Origination (1)", [[HallsOfOrigination\HallsOfOrigination1_]]},
	{6, "Halls Of Origination (2)", [[HallsOfOrigination\HallsOfOrigination2_]]},
	{7, "Halls Of Origination (3)", [[HallsOfOrigination\HallsOfOrigination3_]]},
	{8, "Lost City of Tolvir", [[LostCityofTolvir\LostCityofTolvir]]},
	{9, "The Vortex Pinnacle", [[Skywall\Skywall1_]]},
	{10, "The Bastion of Twilight (1)", [[TheBastionofTwilight\TheBastionofTwilight1_]]},
	{11, "The Bastion of Twilight (2)", [[TheBastionofTwilight\TheBastionofTwilight2_]]},
	{12, "The Bastion of Twilight (3)", [[TheBastionofTwilight\TheBastionofTwilight3_]]},
	{13, "The Deadmines (1)", [[TheDeadmines\TheDeadmines1_]]},
	{14, "The Deadmines (2)", [[TheDeadmines\TheDeadmines2_]]},
	{15, "The Stone Core", [[TheStoneCore\TheStoneCore1_]]},
	{16, "Throne of Tides (1)", [[ThroneOfTides\ThroneOfTides1_]]},
	{17, "Throne of Tides (2)", [[ThroneOfTides\ThroneOfTides2_]]},
	{18, "Throne of the Four Winds", [[Throneofthefourwinds\Throneofthefourwinds1_]]},
	{19, "Icecrown Citadel (1)", [[IcecrownCitadel\IcecrownCitadel1_]]},
	{20, "Icecrown Citadel (2)", [[IcecrownCitadel\IcecrownCitadel2_]]},
	{21, "Icecrown Citadel (3)", [[IcecrownCitadel\IcecrownCitadel3_]]},
	{22, "Icecrown Citadel (4)", [[IcecrownCitadel\IcecrownCitadel4_]]},
	{23, "Icecrown Citadel (5)", [[IcecrownCitadel\IcecrownCitadel5_]]},
	{24, "Icecrown Citadel (6)", [[IcecrownCitadel\IcecrownCitadel6_]]},
	{25, "Icecrown Citadel (7)", [[IcecrownCitadel\IcecrownCitadel7_]]},
	{26, "Icecrown Citadel (8)", [[IcecrownCitadel\IcecrownCitadel8_]]},
	}
local colors = {
	{1.000,0.000,0.000,1},
	{1.000,0.549,0.000,1},
	{1.000,1.000,0.000,1},
	{0.000,0.850,0.000,1},
	{0.000,0.502,0.502,1},
	{0.000,0.000,1.000,1},
	{0.502,0.000,0.502,1},
	{1.000,0.412,0.706,1},
	{0.502,0.502,0.502,1},
	{0.000,0.000,0.000,1},
	{0.545,0.271,0.075,1}
}
local bases = { '0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z' }; -- for rebasing
local OpenDoodles = {};
local dw,dh = 480, 320; -- default width/height. Used for the uniform coordinate system.
local penWidth = 80; -- default pen size.
local myName = UnitName("player"); -- my name...or your name!
local syntax = "S1"; -- Syntax version control
local dpd_debug = false;

Doodlepad.consoleOptions = {
	name = "Doodlepad",
	type = 'group',
	args = {
		confdesc = {
			order = 1,
			type = "description",
			name = "Here you can set the general configuration of Doodlepad.",
			cmdHidden = true
		},
		receivingDoodles = {
			order = 1,
			name = "Receiving doodles",
			type = "group",
			inline = true,
			args = {
				raidRank = {
				order = 1,
				name  = "Receiving from raid channel",
				desc = "Here you can set who will be allowed to |cff44cc22send|r you doodles through the raid channel",
				type = 'select',
				values = {'Everyone','Raid Leader + Assisstant', 'Raid Leader'},
				get = function(info) return Doodleboard.allowReceiveRaid; end,
				set = function(info,v)
					Doodleboard.allowReceiveRaid = v or Doodleboard.allowReceiveRaid;
					end,
				},
				guildRank = {
				order = 1,
				name  = "Receiving from guild channel",
				desc = "Here you can set who will be allowed to |cff44cc22send|r you doodles through the guild channel",
				type = 'select',
				values = {'Guild Master','Rank 1','Rank 2', 'Rank 3', 'Rank 4', 'Rank 5', 'Rank 6', 'Rank 7', 'Rank 8', 'Everyone'},
				get = function(info) return Doodleboard.allowReceiveGuild; end,
				set = function(info,v)
					Doodleboard.allowReceiveGuild = v or Doodleboard.allowReceiveGuild;
					end,
				},
				partyRank = {
				order = 1,
				name  = "Receiving from party channel",
				desc = "Here you can set who will be allowed to |cff44cc22send|r you doodles through the party channel",
				type = 'select',
				values = {'Everyone', 'Party Leader'},
				get = function(info) return Doodleboard.allowReceiveParty; end,
				set = function(info,v)
					Doodleboard.allowReceiveParty = v or Doodleboard.allowReceiveParty;
					end,
				},
				whisperRank = {
				order = 1,
				name  = "Receiving private doodles",
				desc = "Here you can set who will be allowed to |cff44cc22send|r you doodles privately",
				type = 'select',
				values = {'Everyone','Friends only'},
				get = function(info) return Doodleboard.allowReceiveWhisper; end,
				set = function(info,v)
					Doodleboard.allowReceiveWhisper = v or Doodleboard.allowReceiveWhisper;
					end,
				},
			}
		},
		collabDoodles = {
			order = 2,
			name = "Doodle collaboration",
			type = "group",
			inline = true,
			args = {
				raidRank = {
				order = 1,
				name  = "Collaboration in raid channel",
				desc = "Here you can set who will be allowed to |cffcc4422edit|r doodles that others are broadcasting in the raid",
				type = 'select',
				values = {'Everyone','Raid Leader + Assisstant', 'Raid Leader', 'Original author only'},
				get = function(info) return Doodleboard.allowEditRaid; end,
				set = function(info,v)
					Doodleboard.allowEditRaid = v or Doodleboard.allowEditRaid;
					end,
				},
				guildRank = {
				order = 1,
				name  = "Receiving from guild channel",
				desc = "Here you can set who will be allowed to |cffcc4422edit|r doodles that others are broadcasting in the guild",
				type = 'select',
				values = {'Guild Master','Rank 1','Rank 2', 'Rank 3', 'Rank 4', 'Rank 5', 'Rank 6', 'Rank 7', 'Rank 8', 'Everyone', 'Original author only'},
				get = function(info) return Doodleboard.allowEditGuild; end,
				set = function(info,v)
					Doodleboard.allowEditGuild = v or Doodleboard.allowEditGuild;
					end,
				},
				partyRank = {
				order = 1,
				name  = "Receiving from party channel",
				desc = "Here you can set who will be allowed to |cffcc4422edit|r doodles that others are broadcasting in your party",
				type = 'select',
				values = {'Everyone', 'Party Leader', 'Original author only'},
				get = function(info) return Doodleboard.allowEditParty; end,
				set = function(info,v)
					Doodleboard.allowEditParty = v or Doodleboard.allowEditParty;
					end,
				},
			}
		}
	}
}

--[[ CONFIG MENU AND DEFAULTS ]]--
LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("Doodlepad", Doodlepad.consoleOptions)
LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Doodlepad", "Doodlepad")

local function Doodlepad_SetDefaults() -- Set default configuration values
	Doodleboard.allowReceiveRaid = Doodleboard.allowReceiveRaid or 1;
	Doodleboard.allowReceiveGuild = Doodleboard.allowReceiveGuild or 9;
	Doodleboard.allowReceiveWhisper = Doodleboard.allowReceiveWhisper or 1;
	Doodleboard.allowReceiveParty = Doodleboard.allowReceiveParty or 1;
	Doodleboard.allowEditRaid = Doodleboard.allowEditRaid or 2;
	Doodleboard.allowEditGuild = Doodleboard.allowEditGuild or 2;
	Doodleboard.allowEditParty = Doodleboard.allowEditParty or 1;
end


--[[ GENERIC TOOLS ]]--
local function rebase(number, from, to)
	local num = tonumber(number, from) or 0;
	local y,retval,rem = to - 1,"",num;
	while ( num > y ) do local x = floor(num/to); rem = num - (x*to); num = x;retval = (bases[rem+1] or 0) .. (retval or 0);end
	retval = (bases[num+1] or 0) .. ( retval or 0);
	return retval;
end
local function Doodle_Broadcast_Prepare(doodle, cdoodle)
	if ( doodle.chan and doodle.chan:len() > 1 and ((#cdoodle > 1 or (cdoodle[1] and cdoodle[1][4] ~= 0)) or (cdoodle[1] and cdoodle[1].text))) then
		if ( (doodle.chan or ""):len() < 2 or (doodle.uid or ""):len() == 0 or doodle.chan == myName ) then return end
		if ( doodle.chan == "WHISPER" ) then doodle.chan = doodle.sender; end
		local x = 0;
		local c = "";
		if ( not cdoodle[1].text and not cdoodle[1].map ) then
			local color = ("%02x%02x%02x"):format(cdoodle.color[1]*255, cdoodle.color[2]*255, cdoodle.color[3]*255);
			for n = 1, #cdoodle do
				x = x + 1;
				c = c .. ("%2s%2s%1u"):format(rebase(floor(cdoodle[n][1]), 10, 32), rebase(floor(cdoodle[n][2]), 10, 32), cdoodle[n][4] or 0);
				if ( x > 40 ) then
					if ( doodle.chan == "RAID" or doodle.chan == "GUILD" or doodle.chan == "PARTY" or doodle.chan == "BATTLEGROUND" ) then
						tinsert(Doodlepad.messages, {doodle.uid .. syntax .. "D" .. color .. c, doodle.chan});
					else
						tinsert(Doodlepad.messages, {doodle.uid .. syntax .."D" .. color .. c, "WHISPER", doodle.chan});
					end
					c = ("%2s%2s%1u"):format(rebase(floor(cdoodle[n][1]), 10, 32), rebase(floor(cdoodle[n][2]), 10, 32), cdoodle[n][4] or 0);
					x = 0;
				end
			end
			if ( c:len() > 0 ) then
				if ( doodle.chan == "RAID" or doodle.chan == "GUILD" or doodle.chan == "PARTY" or doodle.chan == "BATTLEGROUND" ) then
					tinsert(Doodlepad.messages, {doodle.uid .. syntax .."D" .. color .. c, doodle.chan});
				else
					tinsert(Doodlepad.messages, {doodle.uid .. syntax .."D" .. color .. c, "WHISPER", doodle.chan});
				end
			end
		elseif ( not cdoodle[1].map ) then
			local c = ("%2s%2s"):format(rebase(floor(cdoodle[1].x), 10, 32), rebase(floor(cdoodle[1].y), 10, 32));
			if ( doodle.chan == "RAID" or doodle.chan == "GUILD" or doodle.chan == "PARTY" or doodle.chan == "BATTLEGROUND" ) then
				tinsert(Doodlepad.messages, {doodle.uid .. syntax .."T" .. c .. cdoodle[1].text, doodle.chan});
			else
				tinsert(Doodlepad.messages, {doodle.uid .. syntax .."T" .. c .. cdoodle[1].text, "WHISPER", doodle.chan});
			end
		else
			local c = ("%2s%04u%04u%04u%04u"):format(rebase(floor(cdoodle[1].map), 10, 32), unpack(cdoodle[1].mapCoordsLarge or {0,0,4000,3000}));
			if ( doodle.chan == "RAID" or doodle.chan == "GUILD" or doodle.chan == "PARTY" or doodle.chan == "BATTLEGROUND" ) then
				tinsert(Doodlepad.messages, {doodle.uid .. syntax .."M" .. c, doodle.chan});
			else
				tinsert(Doodlepad.messages, {doodle.uid .. syntax .."M" .. c, "WHISPER", doodle.chan});
			end
		end
	end
end

local function MapOverlay(rec, announce)
	if not rec or not rec.drawLayer then return end
	local map, canvas, x1,y1,x2,y2 = nil, rec.drawLayer, unpack(rec.mapCoords or {0,0,4,3});
	if rec.map then
		rec.map = tonumber(rec.map);
		for n=1,#instanceMaps do
			if instanceMaps[n][1] == rec.map then
				map = instanceMaps[n][3];
				break;
			end
		end
	end
	x1,y1,x2,y2 = x1 or 0,y1 or 0,x2 or 4,y2 or 3; -- default to the full 3x4 tile rectangle
	local _,_,w,h = canvas:GetRect();
	if not canvas.overlays then -- Create the texture objects needed for displaying the maps
		canvas.overlays = {};
		for a = 1, 3 do
			for b = 1, 4 do
				local n = ((a-1)*4) + b;
				canvas.overlays[n] = canvas:CreateTexture();
				local c = canvas.overlays[n];
				c:SetDrawLayer("BACKGROUND")
				if n == 1 then c:SetPoint("TOPLEFT", canvas, "TOPLEFT");
				elseif b == 1 then c:SetPoint("TOPLEFT", canvas.overlays[n-4], "BOTTOMLEFT");
				else c:SetPoint("TOPLEFT", canvas.overlays[n-1], "TOPRIGHT");
				end
			end
		end
	end
	local x, y = x2-x1, y2-y1; -- Total width and height of the map to show
	for a = 1, 3 do
		for b = 1, 4 do
			local n = ((a-1)*4) + b;
			local c = canvas.overlays[n];
			if map then
				c:SetTexture(("Interface\\WorldMap\\%s%u"):format(map, n));
				c:SetAlpha(0.8)
				-- height
				local ty1,ty2 = max(a-1, y1), min(a, y2);
				if ty2 <= ty1 then
					c:SetHeight(0.001)
					ty1, ty2 = 0,0
				else
					local s = ((ty2-ty1)/y)*h;
					c:SetHeight(s);
					ty1,ty2 = ty1-a+1, ty2-a+1;
				end
				-- width
				local tx1,tx2 = max(b-1, x1), min(b, x2);
				if tx2 <= tx1 then
					c:SetWidth(0.001)
					tx1,tx2 = 0,0;
				else
					local s = ((tx2-tx1)/x)*w;
					c:SetWidth(s);
					tx1,tx2 = tx1-b+1, tx2-b+1;
				end
				c:SetTexCoord(tx1,tx2,ty1,ty2);
			else
				canvas.overlays[n]:SetTexture(nil);
			end
		end
	end
	if announce then Doodle_Broadcast_Prepare(rec.doodle, {{map=rec.map, mapCoordsLarge={x1*1000,y1*1000,x2*1000,y2*1000}}}); end
end

local function Doodle_ParseLink(...)
	local linktype, uid = strsplit(":", (select(1, ...)))
	if linktype == "DPD" then
		for n = 1, #OpenDoodles do
			if OpenDoodles[n].uid == uid then OpenDoodles[n].frame:Show() break end
		end
	else
		Doodlepad.itemref(...)
	end
end

local function Doodle_DrawLine(rec, T, sx, sy, ex, ey, color)
	sx = sx * (rec.width/dw);
	sy = -sy * (rec.height/dh);
	ex = ex * (rec.width/dw);
	ey = -ey * (rec.height/dh);
	T:SetTexCoord(0,1,0,1);
	local C = rec.drawLayer;
    local w = penWidth;
    local dx,dy = ex - sx, ey - sy;
    local cx,cy = (sx + ex) / 2, (sy + ey) / 2;
    if (dx < 0) then
        dx,dy = -dx,-dy;
    end
    local Z = (256/255) / 2;
    local l = sqrt((dx * dx) + (dy * dy));
    local s,c = -dy / l, dx / l;
    local sc = s * c;
    local Bwid, Bhgt, BLx, BLy, TLx, TLy, TRx, TRy, BRx, BRy;
    if (dy >= 0) then
        Bwid = ((l * c) - (w * s)) * Z;
        Bhgt = ((w * c) - (l * s)) * Z;
        BLx, BLy, BRy = (w / l) * sc, s * s, (l / w) * sc;
        BRx, TLx, TLy, TRx = 1 - BLy, BLy, 1 - BRy, 1 - BLx;
        TRy = BRx;
    else
        Bwid = ((l * c) + (w * s)) * Z;
        Bhgt = ((w * c) + (l * s)) * Z;
        BLx, BLy, BRx = s * s, -(l / w) * sc, 1 + (w / l) * sc;
        BRy, TLx, TLy, TRy = BLx, 1 - BRx, 1 - BLx, 1 - BLy;
        TRx = TLy;
    end
    T:SetDrawLayer("BACKGROUND", 3)
	T:ClearAllPoints();
	if not (Bwid-1<Bwid) or not (Bhgt-1<Bhgt) then return; end -- discard bad data (IND or INF)
    T:SetVertexColor(color[1],color[2],color[3],color[4]);
    T:SetTexCoord(TLx, TLy, BLx, BLy, TRx, TRy, BRx, BRy);
    T:SetPoint("TOPLEFT",   C, "TOPLEFT", cx - Bwid, cy + Bhgt);
    T:SetSize(Bwid*2,Bhgt*2);
    T:Show()
end

local function channelColor(chan)
	local c = chan:upper():gsub("[^A-Z]", "");
	return ("|cff%02x%02x%02x%s|r"):format(ChatTypeInfo[c].r*255,ChatTypeInfo[c].g*255, ChatTypeInfo[c].b*255, chan);
end


--[[ PERMISSION CHECKS ]]--
local function Doodlepad_AllowReceive(channel, name)
	if channel == "WHISPER" then
		if Doodleboard.allowReceiveWhisper == 1 then return true; -- Everyone is allowed to whisper doodles
		elseif Doodleboard.allowReceiveWhisper == 2 then -- Only friends can whisper
			local numfriends = GetNumFriends()
			if not numfriends or numfriends == 0 then return false end
			for i=1,numfriends do
				local friend = (GetFriendInfo(i))
				if friend and friend == name then
					return true
				end
			end
		end
	elseif channel == "GUILD" then
		local total = GetNumGuildMembers(true)
		for i = 1, total do
			local name, rank, rankIndex = GetGuildRosterInfo(i)
			if name and name == sender then
				if rankIndex and rankIndex <= (Doodleboard.allowReceiveGuild or 9) then
					return true
				end
				return false
			end
		end
	elseif channel == "RAID" or channel == "BATTLEGROUND" then
		return (Doodleboard.allowReceiveRaid == 3 and UnitIsPartyLeader(name)) or (Doodleboard.allowReceiveRaid == 2 and UnitIsRaidOfficer(name)) or (Doodleboard.allowReceiveRaid == 1);
	elseif channel == "PARTY" then
		return (Doodleboard.allowReceiveParty == 2 and UnitIsPartyLeader(name)) or (Doodleboard.allowReceiveParty == 1);
	end
end

function Doodlepad_AllowEdit(channel, name)
	if channel == "WHISPER" then return true
	elseif channel == "GUILD" then
		local total = GetNumGuildMembers(true)
		for i = 1, total do
			local name, rank, rankIndex = GetGuildRosterInfo(i)
			if name and name == sender then
				if rankIndex and rankIndex <= (Doodleboard.allowEditGuild or 9) then
					return true
				end
				return false
			end
		end
	elseif channel == "RAID" or channel == "BATTLEGROUND" then
		return (Doodleboard.allowEditRaid == 3 and UnitIsPartyLeader(name)) or (Doodleboard.allowEditRaid == 2 and UnitIsRaidOfficer(name)) or (Doodleboard.allowEditRaid == 1);
	elseif channel == "PARTY" then
		return (Doodleboard.allowEditParty == 2 and UnitIsPartyLeader(name)) or (Doodleboard.allowEditParty == 1);
	end
end



--[[ DRAWING FUNCTIONS ]]--
local function Doodle_Draw_Point(rec, pointa, pointb, col)
	if ( tonumber(pointa[4]) == 0 ) then
		if ( not pointa[3] and pointb ) then
			local T = tremove(rec.free) or rec.drawLayer:CreateTexture();
			T:SetTexture([[Interface\AddOns\Doodlepad\line]]);
			pointa[3] = T; tinsert(rec.texs, T);
			Doodle_DrawLine(rec, T, pointb[1], pointb[2], pointa[1], pointa[2], col);
		end
	elseif ( tonumber(pointa[4]) ) then
		local x,y = pointa[1], pointa[2];
		x,y = x * (rec.width/dw), y * (rec.height/dh);
		local T = tremove(rec.free) or rec.drawLayer:CreateTexture("OVERLAY", 7);
		T:SetTexture([[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_]] .. pointa[4] or 1);
		T:SetVertexColor(1,1,1,1);
		T:SetWidth(32);T:SetHeight(32);T:SetPoint("TOPLEFT", x-16, -y+16);
		tinsert(rec.texs, T); pointa[3] = T;
		T:Show();
	elseif ( pointa.text ) then
		local x,y = pointa.x, pointa.y;
		x,y = x * (rec.width/dw), y * (rec.height/dh);
		local caption = rec.drawLayer:CreateFontString(nil, "DIALOG");
		caption:SetParent(rec.drawLayer);
		caption:SetFontObject(SystemFont_Shadow_Med2);
		caption:SetTextColor(0.96,0.82,0.2,1);
		caption:SetText(pointa.text);
		caption:SetPoint("CENTER", rec.drawLayer, "TOPLEFT", x, -y);
		tinsert(rec.strings, caption);
	end
end

local function Doodle_Draw(rec, wipe, col)
	if ( wipe == true ) then
		for n = 1, #rec.texs do
			local T = tremove(rec.texs);
			T:Hide();
			T:SetTexCoord(0,1,0,1);
			tinsert(rec.free, T);
		end
		for n = 1, #rec.strings do
			local T = tremove(rec.strings);
			T:Hide();
		end
	else
		for n = 1, #rec.cdoodle do Doodle_Draw_Point(rec, rec.cdoodle[n], rec.cdoodle[n-1], col); end
	end
end

local function Doodle_Redraw(rec)
	Doodle_Draw(rec, true)
	for n = 1, #rec.doodles do
		local doodle = rec.doodles[n];
		local col = doodle.color or colors[5];
		for n = 1, #doodle do doodle[n][3] = nil; Doodle_Draw_Point(rec, doodle[n], doodle[n-1], col); end
	end
	local doodle = rec.cdoodle;
	local col = doodle.color or colors[5];
	for n = 1, #doodle do doodle[n][3] = nil; Doodle_Draw_Point(rec, doodle[n], doodle[n-1], col); end
end

local function Doodlepad_Undo(rec, author, wipe, layer)
	if wipe then
		for n = #rec.doodles, 1, -1 do
			local doodle = rec.doodles[n];
			if ( doodle and (doodle.creator == author or not doodle.creator)) then
				tremove(rec.doodles, n);
			end
		end
		local doodle = rec.doodle or {};
		if ( author == myName and doodle.chan and doodle.uid and doodle.uid:len() == 16) then
			if ( doodle.chan == "RAID" or doodle.chan == "GUILD" or doodle.chan == "PARTY" or doodle.chan == "BATTLEGROUND" ) then
				tinsert(Doodlepad.messages, {doodle.uid .. syntax .."W", doodle.chan});
			elseif ( doodle.chan:len() > 1 ) then
				tinsert(Doodlepad.messages, {doodle.uid .. syntax .."W", "WHISPER", doodle.chan});
			end
		end
	else
		for n = #rec.doodles, 1, -1 do
			local doodle = rec.doodles[n];
			if ( doodle and (doodle.creator == author or not doodle.creator)) then
				if layer then twipe(doodle) else tremove(doodle); end
				if ( #doodle == 0 ) then
					tremove(rec.doodles, n);
				end
				local doodle = rec.doodle or {};
				local T = layer and "L" or "U";
				if ( author == myName and doodle.chan and doodle.uid and doodle.uid:len() == 16) then
					if ( doodle.chan == "RAID" or doodle.chan == "GUILD" or doodle.chan == "PARTY" or doodle.chan == "BATTLEGROUND" ) then
						tinsert(Doodlepad.messages, {doodle.uid .. syntax ..T, doodle.chan});
					elseif ( doodle.chan:len() > 1 ) then
						tinsert(Doodlepad.messages, {doodle.uid .. syntax ..T, "WHISPER", doodle.chan});
					end
				end
				break;
			end
		end
	end
	Doodle_Redraw(rec);
end

local function Doodle_Update(rec,a,b,c,d)
	if ( #rec.cdoodle > 0 ) then
		local l = rec.cdoodle[#rec.cdoodle];
		if ( floor(l[1]) ~= floor(c) or floor(l[2]) ~= floor(d) ) then
			local dist = sqrt( pow(l[1] - c, 2) + pow(l[2] - d, 2));
			if ( dist >= 3 and (rec.x ~= c or rec.y ~= d)) then -- Any lines shorter than 3 pixels get discarded.
				rec.x, rec.y = c,d;
				tinsert(rec.cdoodle, {c*(dw/rec.width),d*(dh/rec.height),nil,0});
			end
		end
	else
		tinsert(rec.cdoodle, {c*(dw/rec.width),d*(dh/rec.height),nil,0});
	end
	Doodle_Draw(rec, nil, rec.cdoodle.color);
end

local function Doodle_Start(rec,a,b,c,d)
	rec.cdoodle.color = rec.color;
	local now = GetTime();
	if ( rec.lastClick and (now - rec.lastClick) < 0.25) then -- double click for placing text
		rec.editbox:ClearAllPoints();
		rec.editbox:SetPoint("CENTER", rec.drawLayer, "TOPLEFT", c, -d);
		rec.editbox:SetWidth(96);
		rec.editbox:SetHeight(16);
		rec.editbox:SetText("");
		rec.editbox.x,rec.editbox.y = c,d;
		rec.editbox:SetJustifyH("CENTER");
		rec.editbox:SetTextInsets(0, 0, 3, 3);
		rec.editbox:Show();
		rec.editbox:Raise();
	end
	rec.lastClick = now;
end

local function Doodle_Stop(rec,a,b,c,d)
	if IsControlKeyDown() then -- Insert raid target icon if control key is down
		tinsert(rec.cdoodle, {c*(dw/rec.width),d*(dh/rec.height), nil, rec.icon or 1});
		Doodle_Draw(rec);
	end
	rec.cdoodle.creator = myName;
	tinsert(rec.doodles, rec.cdoodle);
	rec:StartCapture(Doodlepad.defaultScript);
	Doodle_Broadcast_Prepare(rec.doodle, rec.cdoodle);
	rec.cdoodle = {creator=myName};
end

local function Doodle_Cancel(rec)
	if IsShiftKeyDown() then -- wipe the canvas if shift is down.
		Doodlepad_Undo(rec, myName, true);
	elseif IsControlKeyDown() then
		Doodlepad_Undo(rec, myName, false, true); -- wipe the last layer only (eventually)
	else
		Doodlepad_Undo(rec, myName);
	end
	rec.cdoodle = {creator=myName};
	rec:StartCapture(Doodlepad.defaultScript);
end


--[[ DATA HANDLING ]]--
local function Doodle_Broadcast()
	if ( #Doodlepad.messages > 0 ) then
		local now = GetTime();
		Doodlepad.last = Doodlepad.last or now;
		if ( now - Doodlepad.last >= 0.6 ) then
			Doodlepad.last = now;
			local msg, chan, target = unpack(tremove(Doodlepad.messages, 1));
			SendAddonMessage("DPD2", msg, chan, target);
			if dpd_debug then print(msg) end
		end
	end
end

local function Doodlepad_Incoming(prefix,msg,chan,sender)
	if dpd_debug then sender = "Moomintroll" end
	if ( sender == myName ) then return end
	if ( prefix == "DPD2" ) then
		local uid, syn, typ, data = msg:sub(1,16), msg:sub(17,18), msg:sub(19,19), msg:sub(20);
		if dpd_debug then uid = "1234567890123456" end
		if syn ~= syntax then return end
		local found = false;
		local doodle = nil;
		for n = 1, #OpenDoodles do
			if ( OpenDoodles[n].uid == uid ) then
				found = true;
				doodle = OpenDoodles[n];
				break;
			end
		end
		if not found then
			if ( Doodlepad_AllowReceive(chan, sender) ) then
				local frame, recorder = Doodlepad_window(chan, uid,true,sender);
				frame.caption:SetText("Doodlepad - Broadcast by " .. sender);
				doodle = { frame = frame, recorder = recorder, uid = uid, sender = sender, painters={sender}};
				tinsert(OpenDoodles, doodle );
				doodle.painters = {sender};
				doodle.chan = chan;
				print(string.format("Doodlepad: %s is broadcasting a doodle. |HDPD:%s|h|cFFFFFF00[Click here]|r|h to view it",sender, uid));
			end
		end
		if ( doodle ) then
			if sender == doodle.sender or Doodlepad_AllowEdit(chan, sender) then
				local found = false;
				for i = 1, #doodle.painters do
					if ( doodle.painters[i] == sender ) then found = true; break; end
				end
				if ( not found ) then
					print( ("Doodlepad: %s has joined %s's doodle session"):format(sender, doodle.sender));
					tinsert(doodle.painters, sender);
				end
				local rec = doodle.recorder;
				if ( #rec.cdoodle > 0 ) then
					tinsert(rec.doodles, rec.cdoodle);
					rec.cdoodle = {creator=sender};
				end
				if ( typ == "D" ) then -- drawings
					local color, data = data:sub(1,6), data:sub(7);
					local Color = {}; string.gsub(color, "([0-9a-fA-F][0-9a-fA-F])", function (x) tinsert(Color, tonumber(x, 16) / 255); end);
					rec.cdoodle.creator = sender;
					rec.cdoodle.color = Color;
					for n = 1, floor(data:len()/5) do
						local p = ((n-1)*5)+1;
						local x, y, z = rebase(data:sub(p, p+1), 32, 10), rebase(data:sub(p+2, p+3), 32, 10), data:sub(p+4, p+4);
						tinsert(rec.cdoodle, {x,y,nil,z});
						Doodle_Draw(rec, false, Color);
					end
					tinsert(rec.doodles, rec.cdoodle);
					rec.cdoodle = {creator=sender};
				elseif ( typ == "U" ) then Doodlepad_Undo(rec, sender); -- Undo single step
				elseif ( typ == "L" ) then Doodlepad_Undo(rec, sender, false, true); -- Undo layer
				elseif ( typ == "W" ) then Doodlepad_Undo(rec, sender, true); -- Undo all (wipe)
				elseif ( typ == "T" ) then
					local x, y, t = rebase(data:sub(1,2), 32, 10), rebase(data:sub(3,4), 32, 10), data:sub(5);
					tinsert(rec.cdoodle, {text=t, x=x,y=y});
					Doodle_Draw(rec);
					tinsert(rec.doodles, rec.cdoodle);
					rec.cdoodle = {creator=sender};
				elseif ( typ == "M" ) then
					local m,a,b,c,d = rebase(data:sub(1,2), 32, 10), tonumber(data:sub(3,6)) or 0,tonumber(data:sub(7,10)) or 0,tonumber(data:sub(11,14)) or 4000,tonumber(data:sub(15,18)) or 3000;
					a,b,c,d = a/1000,b/1000,c/1000,d/1000;
					rec.map = m or 0;
					rec.mapCoords = {a,b,c,d};
					MapOverlay(rec);
				end
			end
		end
	end
end

local function Storyboard_Save(doodle)
	local canvas;
	for n = 1, #Doodleboard do
		if ( Doodleboard[n].uid == doodle.uid ) then
			Doodleboard[n].doodles = {};
			canvas = Doodleboard[n].doodles;
			Doodleboard[n].map = doodle.recorder.map;
			Doodleboard[n].mapCoords = doodle.recorder.mapCoords;
			break;
		end
	end
	if ( not canvas ) then
		Doodleboard[#Doodleboard+1] = {uid=doodle.uid, creator = doodle.sender, doodles = {}, map = doodle.recorder.map, mapCoords = doodle.recorder.mapCoords};
		canvas = Doodleboard[#Doodleboard].doodles;
	end
	for n = 1, #doodle.recorder.doodles do
		local doodle = doodle.recorder.doodles[n];
		if ( doodle[1] and doodle[1].text ) then
			local c = ("T%2s%2s%s"):format(rebase(floor(doodle[1].x or 0), 10, 32), rebase(floor(doodle[1].y or 0), 10, 32), doodle[1].text or "");
			tinsert(canvas, c);
		else
			if not doodle.color then doodle.color = {0,0,0} end
			local c = ("D%02x%02x%02x"):format(doodle.color[1]*255, doodle.color[2]*255, doodle.color[3]*255);
			for i = 1, #doodle do
				c = c .. ("%02s%02s%1u"):format(rebase(floor(doodle[i][1] or 0), 10, 32), rebase(floor(doodle[i][2] or 0), 10, 32), doodle[i][4] or 0);
			end
			tinsert(canvas, c);
		end
	end
end

local function Storyboard_Load(uid)
	local doodle;
	local retdoodle = {};
	for n = 1, #Doodleboard do
		if ( Doodleboard[n].uid == uid ) then
			doodle = Doodleboard[n];
			break;
		end
	end
	if (doodle) then
		retdoodle.uid = doodle.uid;
		retdoodle.creator = myName;
		retdoodle.doodles = {};
		retdoodle.cdoodle = {creator=myName};
		retdoodle.map = doodle.map;
		retdoodle.mapCoords = doodle.mapCoords;
		for n = 1, #doodle.doodles do
			local d = doodle.doodles[n];
			local t = d:sub(1,1);
			local data = d:sub(2);
			if ( t == "D" ) then
				local color, data = data:sub(1,6), data:sub(7);
				local Color = {}; string.gsub(color, "([0-9a-fA-F][0-9a-fA-F])", function (x) tinsert(Color, tonumber(x, 16) / 255); end);
				local s = {creator=myName, color=Color};
				for n = 1, floor(data:len()/5) do
					local p = ((n-1)*5)+1;
					local x, y, z = rebase(data:sub(p, p+1), 32, 10), rebase(data:sub(p+2, p+3), 32, 10), data:sub(p+4, p+4);
					tinsert(s, {x,y,nil,z});
				end
				tinsert(retdoodle.doodles, s);
			elseif (t == "T" ) then
				local x, y, t = rebase(data:sub(1,2), 32, 10), rebase(data:sub(3,4), 32, 10), data:sub(5);
				tinsert(retdoodle.doodles, {{x=x,y=y,text=t}});
			end
		end
	end
	return retdoodle;
end

local function Storyboard_Scroll(value)
	Storyboard_Frame.X = (Storyboard_Frame.X or 1) - value;
	if ( (Storyboard_Frame.X+2) > #Doodleboard ) then Storyboard_Frame.X = #Doodleboard - 2; end
	if ( Storyboard_Frame.X < 1 ) then Storyboard_Frame.X = 1; end
	local i = 0;
	for n = Storyboard_Frame.X, Storyboard_Frame.X+2 do
		i = i + 1;
		if ( Doodleboard[n] ) then
			Doodleboard[n].title = Doodleboard[n].title or "Unnamed Doodle";
			local doodle = Storyboard_Load(Doodleboard[n].uid);
			Storyboard_Frame.doodles[i].doodles = doodle.doodles;
			Storyboard_Frame.doodles[i].uid = doodle.uid;
			Storyboard_Frame.doodles[i].map = doodle.map;
			Storyboard_Frame.doodles[i].mapCoords = doodle.mapCoords;
			Storyboard_Frame.doodles[i].creator = doodle.creator;
			Storyboard_Frame.doodles[i].caption:SetText(Doodleboard[n].title);
			Storyboard_Frame.doodles[i].pointer = Doodleboard[n];
			Doodle_Redraw(Storyboard_Frame.doodles[i]);
			MapOverlay(Storyboard_Frame.doodles[i]);
			Storyboard_Frame.doodles[i].drawLayer:SetScript("OnEnter", function(caller)
			GameTooltip:ClearLines();
			GameTooltip:SetOwner(caller, "ANCHOR_BOTTOM");
			GameTooltip:SetText(Doodleboard[n].title);
			GameTooltip:AddLine("Created by " .. (doodle.creator or "??"));
			GameTooltip:Show();
			end);
			Storyboard_Frame.doodles[i].drawLayer:SetScript("OnLeave", function() GameTooltip:Hide() end);
		else
			Storyboard_Frame.doodles[i].doodles = {};
			Storyboard_Frame.doodles[i].uid = nil;
			Storyboard_Frame.doodles[i].map = nil;
			Storyboard_Frame.doodles[i].creator = myName;
			Storyboard_Frame.doodles[i].caption:SetText("");
			Storyboard_Frame.doodles[i].drawLayer:SetScript("OnEnter",nil);
			Storyboard_Frame.doodles[i].drawLayer:SetScript("OnLeave",nil);
			Doodle_Redraw(Storyboard_Frame.doodles[i]);
			Doodle_Redraw(Storyboard_Frame.doodles[i]);
		end
	end
end

local function Storyboard_Delete(uid)
	for n = 1, #Doodleboard do
		if ( Doodleboard[n].uid == uid ) then
			tremove(Doodleboard, n);
			break;
		end
	end
	Storyboard_Scroll(0);
end



--[[ UI FUNCTIONS ]]--
local function Doodlepad_CreateColorBtn(parentFrame)
	local button = CreateFrame("Button",nil,parentFrame)
	button:SetWidth(30)
	button:SetHeight(30)
	button:SetPoint("BOTTOMLEFT", 5, 3)

	button:SetNormalTexture([[Interface\ChatFrame\ChatFrameColorSwatch]])
	button:RegisterForClicks("AnyUp")
	button:SetScript("OnClick",
		function(self, mbutton, down)
			if ColorPickerFrame:IsShown() then
				ColorPickerFrame:Hide()
			else
				button.r, button.g, button.b = unpack(parentFrame.recorder.color);
				button.hasOpacity = false
				button.swatchFunc =
					function()
						button.r, button.g, button.b = ColorPickerFrame:GetColorRGB()
						button:GetNormalTexture():SetVertexColor(button.r, button.g, button.b)
						parentFrame.recorder.color = {button.r, button.g, button.b};
					end
				button.cancelFunc =
					function()
						button.r, button.g, button.b = ColorPicker_GetPreviousValues()
						button:GetNormalTexture():SetVertexColor(button.r, button.g, button.b)
						parentFrame.recorder.color = {button.r, button.g, button.b};
					end
				OpenColorPicker(button)
				ColorPickerFrame:SetFrameStrata("TOOLTIP")
				ColorPickerFrame:Raise()
			end
		end)
	button:GetNormalTexture():SetVertexColor(1,0,0)
	return button
end

function Doodlepad_window(chan, uid, silent, sender)
	uid = uid or "";
    local frame = CreateFrame("Frame", nil, UIParent);
    frame:SetBackdrop({bgFile = "Interface/GLUES/loading",edgeFile = "Interface/GLUES/Common/TextPanel-Border",edgeSize = 16,insets = { left = 4, right = 4, top = 4, bottom = 4 }});
    frame:SetBackdropColor(0.3,0.3,0.3,0.85);
    frame:SetWidth(dw)
    frame:SetHeight(dh);
    frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
	frame:RegisterForDrag("LeftButton");
	frame:SetScript("OnDragStart",function()  frame:StartMoving(); end);
	frame:SetScript("OnDragStop", function()  frame:StopMovingOrSizing(); end);
	frame:SetMovable(true);
	frame:EnableMouse(true);
	frame:EnableMouseWheel(true);
	frame:SetResizable(true);
	local subFrame = CreateFrame("Frame");
	subFrame:SetParent(frame);
	subFrame:SetPoint("TOPLEFT", 5, -28);
	subFrame:SetPoint("BOTTOMRIGHT", -5, 32);
	local canvas = subFrame:CreateTexture("BACKGROUND");
	canvas:SetTexture(1,1,1,0.85);
	canvas:SetAllPoints(subFrame);
	local recorder = libmg:New(subFrame);
	frame.recorder = recorder;
	recorder.colorbtn = Doodlepad_CreateColorBtn(frame);
	local icon = CreateFrame("Frame", nil, frame)
	local tex = icon:CreateTexture("ARTWORK");
	tex:SetTexture([[Interface\Icons\Ability_Spy]])
	tex:SetAllPoints(icon);
	icon:SetSize(24,24);
	icon:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 36,5);
	icon:EnableMouse(true);
	icon:SetScript("OnMouseUp", function(frame) myRec = recorder; overlaySelector:Show(); overlaySelector:Raise(); end);

	local xframe = subFrame;
	if ( uid:len() == 16 ) then frame:Hide();
	else uid = ("%08x%08x"):format(random(time()), random(GetTime())); end
	recorder.editbox = CreateFrame("EditBox", ("DoodleEditbox_%08x"):format(random(time())), recorder.drawLayer, "InputBoxTemplate");
    recorder.editbox:SetAutoFocus(true);
    recorder.editbox:SetScript("OnEscapePressed", function(me) me:Hide(); end);
    recorder.editbox:SetScript("OnEnterPressed",  function(me)
		local txt = {text=me:GetText(), x = me.x*(dw/recorder.width),y=me.y*(dh/recorder.height)};
		tinsert(recorder.cdoodle, txt);
		Doodle_Draw(recorder);
		tinsert(recorder.doodles, recorder.cdoodle);
		Doodle_Broadcast_Prepare(recorder.doodle, recorder.cdoodle);
		recorder.cdoodle = {creator=myName};
		me:Hide();
		end);
    recorder.editbox:SetTextInsets(0, 0, 3, 3);
    recorder.editbox:SetMaxLetters(100);
    recorder.editbox:Hide();

	-- caption
	local caption = frame:CreateFontString(nil, "DIALOG");
	caption:SetParent(frame);
	caption:SetFontObject(ChatFontNormal);
	caption:SetText("Doodlepad");
	caption:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5);
	caption:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", 5, -25);
	frame.caption = caption;

    -- close button
    local close = CreateFrame("button", nil, frame, "UIPanelCloseButton");
    close:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, 0);
    close:Show();

	-- resizer
    local sizer = CreateFrame("frame", nil, frame);
    sizer:SetPoint("BOTTOMRIGHT")
    sizer:SetHeight(15);
    sizer:SetWidth(15);
    sizer:EnableMouse()
    sizer:SetScript("OnMouseDown",function(frame) frame:GetParent():StartSizing("BOTTOMRIGHT") subFrame:SetAlpha(0.25) end)
    sizer:SetScript("OnMouseUp", function(frame) frame:GetParent():StopMovingOrSizing() subFrame:SetAlpha(1) recorder.width = xframe:GetWidth(); recorder.height = xframe:GetHeight(); Doodle_Redraw(recorder); end)
	frame:SetScript("OnSizeChanged", function(frame) MapOverlay(recorder, false);end);

    local line1 = sizer:CreateTexture(nil, "BACKGROUND")
    line1:SetWidth(14)
    line1:SetHeight(14)
    line1:SetPoint("BOTTOMRIGHT", 0, 0)
    line1:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border")
    local x = 0.1 * 14/17
    line1:SetTexCoord(0.05 - x, 0.5, 0.05, 0.5 + x, 0.05, 0.5 - x, 0.5 + x, 0.5)

    local line2 = sizer:CreateTexture(nil, "BACKGROUND")
    line2:SetWidth(8)
    line2:SetHeight(8)
    line2:SetPoint("BOTTOMRIGHT", 0,0)
    line2:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border")
    local x = 0.1 * 8/17
    line2:SetTexCoord(0.05 - x, 0.5, 0.05, 0.5 + x, 0.05, 0.5 - x, 0.5 + x, 0.5)

	recorder.width, recorder.height = subFrame:GetSize();
	recorder.doodles, recorder.cdoodle, recorder.texs, recorder.free, recorder.icon,recorder.strings = {},{creator=myName},{},{}, 1, {};
	recorder.color = colors[1];
	recorder.target = recorder.drawLayer:CreateTexture("OVERLAY",7);
	recorder.target:SetTexture([[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_]] .. recorder.icon);
	recorder.target:SetWidth(32);
	recorder.target:SetHeight(32);

	frame:SetScript("OnMouseWheel",
		function(caller, value)
			if ( IsControlKeyDown() ) then
				recorder.icon = recorder.icon + value;
				if ( recorder.icon > 8 ) then recorder.icon = 1; end
				if ( recorder.icon < 1 ) then recorder.icon = 8; end
			else
				recorder.xcolor = (recorder.xcolor or 1) - value;
				if ( recorder.xcolor > #colors ) then recorder.xcolor = 1; end
				if ( recorder.xcolor < 1 ) then recorder.xcolor = #colors; end
				recorder.color = colors[recorder.xcolor];
				recorder.colorbtn:GetNormalTexture():SetVertexColor(unpack(recorder.color))
			end
		end);
	subFrame:SetScript("OnUpdate",
		function()
			if ( subFrame:IsMouseOver() ) then
				if ( IsControlKeyDown() ) then
					recorder.target:SetTexture([[Interface\TARGETINGFRAME\UI-RaidTargetingIcon_]] .. recorder.icon);
					recorder.target:SetWidth(32);
					recorder.target:SetHeight(32);
				else
					if type(recorder.color) == "number" then
						recorder.target:SetTexture(unpack(colors[recorder.color or 1]));
					else
						recorder.target:SetTexture(unpack(recorder.color));
					end
					recorder.target:SetWidth(4);
					recorder.target:SetHeight(4);
				end
				recorder.target:SetPoint("CENTER", recorder.drawLayer, "TOPLEFT", recorder.mouseEnd[1], -recorder.mouseEnd[2]);
				recorder.target:Show();
			else
				recorder.target:Hide();
			end
		end);
	local xchan = string.upper(chan or "");
	if ( xchan == "RAID" or xchan == "PARTY" or xchan == "GUILD" or xchan == "BATTLEGROUND" ) then
		if not silent then print("Sharing doodle with the " .. chan); caption:SetText("Doodlepad: Broadcasting to the " .. chan); end
		chan = xchan;
	elseif ( xchan:len() > 1 ) then
		if not silent then print("Sharing doodle with " .. chan); caption:SetText("Doodlepad: Sharing with " .. chan); end
	end
	local doodle = nil;
	local found = false;
	for n = 1, #OpenDoodles do
		if OpenDoodles[n].uid == uid then found = true doodle = OpenDoodles[n]; if xchan:len() > 1 then doodle.chan = chan; end break end
	end
	if not found then
		doodle = {uid=uid, recorder=recorder, sender=sender or myName, painters={sender or myName}, frame=frame, chan=chan};
		tinsert(OpenDoodles, doodle);
	end
	close:SetScript("OnClick",
		function()
		frame:Hide();
		Doodle_Draw(recorder, true); -- wipe the pad
		if ( #doodle.recorder.doodles > 0 ) then
			print("Doodlepad closed. The doodle has been saved in your storyboard for later use.");
			Storyboard_Save(doodle);
		end
		Storyboard_Scroll(0);
		end);
	recorder.doodle = doodle;
	recorder:StartCapture(Doodlepad.defaultScript);
	MapOverlay(recorder);
	return frame, recorder;
end

local function Doodle_Open(caller, rec, target)
	EasyMenu({}, Storyboard_Frame.menuFrame, "cursor", 0 , 0, "MENU");
	local f,r = Doodlepad_window(target or "",rec.uid);
	r.uid = rec.uid;
	r.doodles = rec.doodles;
	r.map = rec.map;
	r.mapCoords = rec.mapCoords;
	MapOverlay(r,true)
	Doodle_Redraw(r);
	f:Show();
	f:Raise();
	Storyboard_Frame:Hide();
	if target then
		for n=1, #rec.doodles do Doodle_Broadcast_Prepare(r, rec.doodles[n]); end
	end
end

local function Doodle_OpenTarget(caller, rec)
		EasyMenu({}, Storyboard_Frame.menuFrame, "cursor", 0 , 0, "MENU");
		local frame = CreateFrame("Frame", nil, UIParent);
		frame:SetBackdrop({bgFile = "Interface/GLUES/loading",edgeFile = "Interface/GLUES/Common/TextPanel-Border",edgeSize = 16,insets = { left = 4, right = 4, top = 4, bottom = 4 }});
		frame:SetBackdropColor(0.2,0.2,0.2,0.75);
		frame:SetSize(300,80);
		frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
		frame:EnableMouse(true);
		local editbox = CreateFrame("EditBox", "DoodleWhisper" .. time(), frame, "InputBoxTemplate");
		editbox:SetSize(100,24);
		editbox:SetAutoFocus(true);
		editbox:SetScript("OnEscapePressed", function(me) frame:Hide(); end);
		editbox:SetScript("OnEnterPressed",  function(me)
				local txt = me:GetText();
				Doodle_Open(nil,rec,txt);
				frame:Hide();
				end);
		editbox:SetTextInsets(0, 0, 3, 3);
		editbox:SetMaxLetters(12);
		editbox:SetPoint("CENTER", frame);
		editbox:Show()
		local caption = frame:CreateFontString(nil, "DIALOG");
		caption:SetParent(frame);
		caption:SetFontObject(ChatFontNormal);
		caption:SetText("Enter name of recipient:");
		caption:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5);
		caption:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", 5, -25);
		local close = CreateFrame("button", nil, frame, "UIPanelCloseButton");
		close:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, 0);
		close:Show();
		frame:SetFrameStrata("FULLSCREEN");
		frame:Raise();
	end

local function Storyboard_Context(rec)
	local menu = {
    { text = "Tools:", isTitle = true},
    { text = "Open doodle", func = function() local f,r = Doodlepad_window("", rec.uid); r.chan = ""; r.uid = rec.uid; r.map = rec.map; r.mapCoords = rec.mapCoords; MapOverlay(r) r.doodles = rec.doodles; Doodle_Redraw(r); f:Show(); f:Raise(); Storyboard_Frame:Hide(); end },
	{ text = "Open and broadcast to:", hasArrow = true, menuList = {
		{ text = channelColor("Whisper..."), func = Doodle_OpenTarget, arg1=rec},
        { text = channelColor("Raid"), func = Doodle_Open, arg1=rec, arg2="raid" },
		{ text = channelColor("Party"), func = Doodle_Open, arg1=rec, arg2="party"},
		{ text = channelColor("Guild"), func = Doodle_Open, arg1=rec, arg2="guild"},
		{ text = channelColor("Battleground"), func = Doodle_Open, arg1=rec, arg2="battleground"},
		}
    },
	{ text = "Open as new doodle", func = function() local f,r = Doodlepad_window(""); r.uid = nil; r.doodles = rec.doodles; Doodle_Redraw(r); f:Show(); f:Raise(); Storyboard_Frame:Hide(); end },
	{ text = "Rename", func = function() rec.editbox:SetText(rec.pointer.title); rec.editbox:SetAllPoints(rec.caption); rec.editbox:Show(); end},
	{ text = "Delete doodle", func = function() Storyboard_Delete(rec.uid) end },
    { text = "Cancel"},
    };
    EasyMenu(menu, Storyboard_Frame.menuFrame, "cursor", 0 , 0, "MENU");
end


---- Storyboard ----
Storyboard_Frame = CreateFrame("Frame", nil, UIParent);
Storyboard_Frame:SetBackdrop({bgFile = "Interface/GLUES/loading",edgeFile = "Interface/GLUES/Common/TextPanel-Border",edgeSize = 16,insets = { left = 4, right = 4, top = 4, bottom = 4 }});
Storyboard_Frame:SetBackdropColor(1,1,1,0.4);
Storyboard_Frame:SetWidth(640)
Storyboard_Frame:SetHeight(220);
Storyboard_Frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
Storyboard_Frame:RegisterForDrag("LeftButton");
Storyboard_Frame:SetScript("OnDragStart",function()  Storyboard_Frame:StartMoving(); end);
Storyboard_Frame:SetScript("OnDragStop", function()  Storyboard_Frame:StopMovingOrSizing(); end);
Storyboard_Frame:SetScript("OnMouseWheel", function(caller, value)  Storyboard_Scroll(value); end);
Storyboard_Frame:SetMovable(true);
Storyboard_Frame:EnableMouse(true);
Storyboard_Frame:EnableMouseWheel(true);
Storyboard_Frame:Hide();
Storyboard_Frame.menuFrame = CreateFrame("Frame", "Storyboard_Frame_Menu", UIParent, "UIDropDownMenuTemplate")

--Container frames
Storyboard_Frame.doodles = {};
local x = 10;
for n = 1, 3 do
	Storyboard_Frame.doodles[n] = CreateFrame("Frame", nil, Storyboard_Frame);
	Storyboard_Frame.doodles[n]:SetWidth(200);Storyboard_Frame.doodles[n]:SetHeight(150);
	Storyboard_Frame.doodles[n].canvas = Storyboard_Frame.doodles[n]:CreateTexture("BACKGROUND", -7);
	Storyboard_Frame.doodles[n].drawLayer = CreateFrame("Frame", nil, Storyboard_Frame.doodles[n])
	Storyboard_Frame.doodles[n].drawLayer:SetAllPoints();
	Storyboard_Frame.doodles[n].canvas:SetTexture(1,1,1,1);
	Storyboard_Frame.doodles[n].canvas:SetAllPoints();
	Storyboard_Frame.doodles[n]:SetPoint("TOPLEFT", x,-40);
	Storyboard_Frame.doodles[n].texs = {};
	Storyboard_Frame.doodles[n].free = {};
	Storyboard_Frame.doodles[n].strings = {};
	Storyboard_Frame.doodles[n].cdoodle = {};
	Storyboard_Frame.doodles[n].doodles = {};
	Storyboard_Frame.doodles[n].width = 200;
	Storyboard_Frame.doodles[n].height = 150;
	Storyboard_Frame.doodles[n].drawLayer:EnableMouse(true);
	Storyboard_Frame.doodles[n].drawLayer:SetScript("OnMouseup", function(caller,event) if ( event == "RightButton" ) then Storyboard_Context(Storyboard_Frame.doodles[n]) end end);
	Storyboard_Frame.doodles[n].caption = Storyboard_Frame.doodles[n]:CreateFontString(nil, "DIALOG");
	Storyboard_Frame.doodles[n].caption:SetParent(Storyboard_Frame);
	Storyboard_Frame.doodles[n].caption:SetFontObject(ChatFontNormal);
	Storyboard_Frame.doodles[n].caption:SetText("");
	Storyboard_Frame.doodles[n].caption:SetTextColor(1,0.6,0,0.8);
	Storyboard_Frame.doodles[n].caption:SetPoint("TOPLEFT", Storyboard_Frame.doodles[n], "TOPLEFT", 5, 16);
	Storyboard_Frame.doodles[n].caption:SetPoint("BOTTOMRIGHT", Storyboard_Frame.doodles[n], "TOPRIGHT", -5, -2);
	Storyboard_Frame.doodles[n].editbox = CreateFrame("EditBox", nil, Storyboard_Frame.doodles[n], "InputBoxTemplate");
    Storyboard_Frame.doodles[n].editbox:SetAutoFocus(true);
    Storyboard_Frame.doodles[n].editbox:SetScript("OnEscapePressed", function(me) me:Hide(); end);
    Storyboard_Frame.doodles[n].editbox:SetScript("OnEnterPressed",  function(me)
		local txt = me:GetText();
		Storyboard_Frame.doodles[n].pointer.title = txt;
		me:Hide();
		Storyboard_Scroll(0);
		end);
    Storyboard_Frame.doodles[n].editbox:SetTextInsets(0, 0, 3, 3);
    Storyboard_Frame.doodles[n].editbox:SetMaxLetters(100);
	Storyboard_Frame.doodles[n].editbox:SetJustifyH("CENTER");
    Storyboard_Frame.doodles[n].editbox:Hide();
	x = x + 210;
end

-- caption
local caption = Storyboard_Frame:CreateFontString(nil, "DIALOG");
caption:SetParent(Storyboard_Frame);
caption:SetFontObject(ChatFontNormal);
caption:SetText("Storyboard");
caption:SetPoint("TOPLEFT", Storyboard_Frame, "TOPLEFT", 5, -5);
caption:SetPoint("BOTTOMRIGHT", Storyboard_Frame, "TOPRIGHT", 5, -25);

caption = Storyboard_Frame:CreateFontString(nil, "DIALOG");
caption:SetParent(Storyboard_Frame);
caption:SetFontObject(ChatFontNormal);
caption:SetText("Use your mouse wheel to scroll through the saved doodles. Right click a doodle for file actions.");
caption:SetPoint("TOPLEFT", Storyboard_Frame, "BOTTOMLEFT", 5, 25);
caption:SetPoint("BOTTOMRIGHT", Storyboard_Frame, "BOTTOMRIGHT", 5, 5);

-- close button
local close = CreateFrame("button", nil, Storyboard_Frame, "UIPanelCloseButton");
close:SetPoint("TOPRIGHT", Storyboard_Frame, "TOPRIGHT", 0, 0);
close:Show();


-- Instance map overlay selector
local frame = CreateFrame("Frame", nil, UIParent);
frame:SetBackdrop({bgFile = "Interface/GLUES/loading",edgeFile = "Interface/GLUES/Common/TextPanel-Border",edgeSize = 16,insets = { left = 4, right = 4, top = 4, bottom = 4 }});
frame:SetBackdropColor(0,0,0,0.85);
frame:SetSize(480,420);
frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
frame:RegisterForDrag("LeftButton");
frame:SetScript("OnDragStart",function()  frame:StartMoving(); end);
frame:SetScript("OnDragStop", function()  frame:StopMovingOrSizing(); end);
frame:SetMovable(true);
frame:EnableMouse(true);
frame:EnableMouseWheel(true);
frame:SetResizable(true);
local subFrame = CreateFrame("Frame");
subFrame:SetParent(frame);
subFrame:SetPoint("TOPLEFT", 5, -28);
subFrame:SetPoint("BOTTOMRIGHT", -5, 50);
local canvas = subFrame:CreateTexture("BACKGROUND");
canvas:SetTexture(1,1,1,0.85);
canvas:SetAllPoints(subFrame);
local recorder = libmg:New(subFrame);
frame.recorder = recorder;
overlaySelector = frame;
frame:SetFrameStrata("FULLSCREEN_DIALOG");
-- caption
local caption = frame:CreateFontString(nil, "DIALOG");
caption:SetParent(frame);
caption:SetFontObject(ChatFontNormal);
caption:SetText("Select background:");
caption:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5);
caption:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", 5, -25);
local caption = frame:CreateFontString(nil, "DIALOG");
caption:SetParent(frame);
caption:SetFontObject(ChatFontNormal);
caption:SetText("Use your mouse wheel to scroll through backgrounds.|n|cff44AA55Optionally, you can draw a rectangle on the map to zoom in.|r");
caption:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", 5, 48);
caption:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -5, 5);
frame.caption = caption;
frame:Hide();
-- close button
local close = CreateFrame("button", nil, frame, "UIPanelCloseButton");
close:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, 0);
close:Show();
close:SetScript("OnClick", function() MapOverlay(myRec,true); frame:Hide(); end);

recorder.width, recorder.height = dw,dh;
recorder.target = recorder.drawLayer:CreateTexture("OVERLAY",7);
recorder.target:SetTexture(0.5,0,0.7,0.5);

frame:SetScript("OnMouseWheel",
	function(caller, value)
		recorder.xmap = (recorder.xmap or 1) - value;
		if ( recorder.xmap > #instanceMaps ) then recorder.xmap = 1; end
		if ( recorder.xmap < 1 ) then recorder.xmap = #instanceMaps; end
		recorder.map = instanceMaps[recorder.xmap][1];
		MapOverlay(recorder);
		myRec.map = recorder.map;
	end);

local overlayScript = {
	startButton		= "LeftButtonDown",
	stopButton		= "LeftButtonUp",
	cancelButton    = "RightButtonUp",
	updateFunc		= function(rec)
	local a,b,dx,dy = rec:GetBounds(true);
	local C = rec.drawLayer;
    local T = C.mouseLine or C:CreateTexture()
	C.mouseLine = T;
	T:SetTexture(1,1,1,1);
	T:SetTexCoord(0,1,0,1);
    T:SetDrawLayer(layer or "OVERLAY")
    T:SetVertexColor(0.7,0,0.8,0.5);
    T:SetPoint("TOPLEFT", C, "TOPLEFT", a, -b);
	T:SetSize(dx,dy);
    T:Show()
	end,
	stopFunc 		= function(rec) local _,_,w,h = rec.drawLayer:GetRect(); local a,b,dx,dy = rec:GetBounds(true) a,b = a/(w/4), b/(h/3); myRec.mapCoords = {a,b,a+(dx/(w/4)),b+(dy/(h/3))}; rec:StartCapture(rec._CaptureScript) end,
	cancelFunc		= function(rec) if rec.drawLayer.mouseLine then rec.drawLayer.mouseLine:Hide() end myRec.mapCoords = {0,0,4,3} rec:StartCapture(rec._CaptureScript) end,
}

recorder:StartCapture(overlayScript);

Doodlepad.defaultScript = {
	startButton		= "LeftButtonDown",
	stopButton		= "LeftButtonUp",
	cancelButton    = "RightButtonUp",
	startFunc 		= Doodle_Start,
	updateFunc		= Doodle_Update,
	stopFunc 		= Doodle_Stop,
	cancelFunc		= Doodle_Cancel,
};

SLASH_DOODLE1 = '/doodlepad';
SLASH_DOODLE2 = '/pad';
SLASH_STORYBOARD1 = '/storyboard';
SLASH_STORYBOARD2 = '/board';
SlashCmdList['DOODLE'] = function(chan) Doodlepad_window(chan) end;
SlashCmdList['STORYBOARD'] = function(chan) Storyboard_Scroll(0); Storyboard_Frame:Show(); end;
local frame = CreateFrame("Frame");
frame:SetParent(UIParent);frame:SetPoint("TOPLEFT");
frame:RegisterEvent("CHAT_MSG_ADDON");
frame:RegisterEvent("ADDON_LOADED");
frame:SetScript("OnEvent", function(caller, event, ...) if ( event == "CHAT_MSG_ADDON" ) then Doodlepad_Incoming(...) elseif ( event == "ADDON_LOADED" ) then Doodlepad_SetDefaults(); end end);
frame:SetScript("OnUpdate", Doodle_Broadcast);
Doodlepad.itemref = _G.SetItemRef;
_G.SetItemRef = Doodle_ParseLink;
